import 'package:flutter/material.dart';
import 'package:mobile_ui/src/controls/spacing.dart';

typedef ValueValidator<T> = String? Function(
    T value, String? errorCode, String? errorMsg);
enum ControlPosition { leading, trailing }
enum FieldOrientation { vertical, horizontal }

class FormInputBackgroundStyle {
  const FormInputBackgroundStyle(
      {this.backgroudColor, this.borderColor, this.borderWidth});
  final Color? backgroudColor;
  final Color? borderColor;
  final double? borderWidth;
}

const FormInputBackgroundStyle defalutFormInputBackgroundStyle =
    FormInputBackgroundStyle(backgroudColor: Colors.white);

class FormInputBase<T> extends FormField<T> {
  final bool? isRequired;
  final String? labelText;
  final String? hintText;
  final EdgeInsetsGeometry contentPadding;
  final EdgeInsetsGeometry margin;
  final double spacingFlactor;
  final bool enabled;
  final bool readOnly;
  final TextAlign labelAlign;
  final FieldOrientation orientation;
  final ControlPosition controlPosition;
  final ValueNotifier<T>? valueController;
  final Icon? prefixIcon;
  final bool? isDense;

  FormInputBase({
    Key? key,
    required FormFieldBuilder<T> builder,
    ValueValidator<T?>? validator,
    T? initialValue,
    this.valueController,
    this.prefixIcon,
    this.isRequired,
    this.labelText,
    this.hintText,
    this.contentPadding = const EdgeInsets.all(8),
    this.margin = const EdgeInsets.symmetric(vertical: 8),
    this.spacingFlactor = 2,
    required bool enabled,
    this.isDense,
    bool? readOnly,
    this.labelAlign = TextAlign.left,
    this.orientation = FieldOrientation.vertical,
    //Only valid when the direction is equal to horizontal
    this.controlPosition = ControlPosition.trailing,
    AutovalidateMode? autovalidateMode,
    FormFieldSetter<T>? onSaved,
  })  : assert(initialValue == null || valueController == null),
        enabled = enabled,
        readOnly = (enabled == false || readOnly == true),
        super(
            builder: builder,
            initialValue: initialValue,
            onSaved: onSaved,
            enabled: enabled,
            autovalidateMode: autovalidateMode,
            validator: (value) {
              String? errorCode;
              String? errorMsg;
              final isNullOrEmpty = (value == null ||
                  (value is String && value.isEmpty) ||
                  (value is List && value.isEmpty));
              if (isNullOrEmpty) {
                if (isRequired == true) {
                  errorCode = 'required';
                  errorMsg = '${labelText ?? "This field"} is required.';
                }
              }
              if (validator != null)
                return validator(value, errorCode, errorMsg);
              return errorMsg;
            });

  Widget buildField(FormFieldState<T> field, Widget child,
      {FormInputBackgroundStyle? backgroudStyle}) {
    final theme = Theme.of(field.context);
    if (orientation == FieldOrientation.vertical) {
      return buildVField(
          theme, buildDecorator(field, child, backgroundStyle: backgroudStyle));
    } else {
      return buildDecorator(field, buildHField(theme, child),
          backgroundStyle: backgroudStyle);
    }
  }

  Widget buildHField(ThemeData theme, Widget child) {
    var children = <Widget>[child];
    if (labelText != null && labelText!.isNotEmpty) {
      if (controlPosition == ControlPosition.leading) {
        children.add(Expanded(child: buildLabelText(theme)));
      } else {
        children.insert(0, Expanded(child: buildLabelText(theme)));
      }
    }
    return Padding(
        padding: margin,
        child: Column(crossAxisAlignment: CrossAxisAlignment.start, children: [
          Row(
            children: children,
          )
        ]));
  }

  Widget buildVField(ThemeData theme, Widget child) {
    var children = <Widget>[];
    if (labelText != null && labelText!.isNotEmpty) {
      children.addAll([
        buildLabelText(theme),
        Spacing(spacingFlactor),
      ]);
    }
    return Padding(
        padding: margin,
        child: Column(
            crossAxisAlignment: CrossAxisAlignment.stretch,
            children: [...children, child]));
  }

  Widget buildLabelText(ThemeData theme) {
    return Text(
      labelText ?? '',
      textAlign: labelAlign,
      style: enabled
          ? theme.textTheme.subtitle1
          : theme.textTheme.subtitle1!.copyWith(color: theme.disabledColor),
    );
  }

  Widget buildText(ThemeData theme, String? text, {Color? color}) {
    return Text(
      text ?? '',
      textAlign: TextAlign.left,
      style: enabled
          ? (color == null
              ? theme.textTheme.subtitle1
              : theme.textTheme.subtitle1!.copyWith(color: color))
          : theme.textTheme.subtitle1!.copyWith(color: theme.disabledColor),
    );
  }

  Widget buildIcon(ThemeData theme, IconData icon,
      {GestureTapCallback? onTap, Color? enabledColor}) {
    final color =
        enabled ? enabledColor ?? theme.primaryColor : theme.disabledColor;
    final iconWidget = Icon(
      icon,
      color: color,
    );
    if (onTap != null && !readOnly) {
      return InkWell(
        onTap: onTap,
        child: iconWidget,
      );
    }
    return iconWidget;
  }

  bool getIsNullOrEmpty(T? value) {
    return (value == null ||
        (value is String && value.isEmpty) ||
        (value is List && value.isEmpty));
  }

  Widget buildDecorator(FormFieldState<T> field, Widget child,
      {FormInputBackgroundStyle? backgroundStyle}) {
    final effectiveDecoration =
        InputDecoration(contentPadding: EdgeInsets.all(0)).copyWith(
            hintText: hintText,
            border: OutlineInputBorder(borderSide: BorderSide.none));
    final isNullOrEmpty = getIsNullOrEmpty(field.value);

    if (backgroundStyle == null) {
      return InputDecorator(
          decoration: effectiveDecoration.copyWith(
              hintText: hintText, errorText: field.errorText, isDense: isDense),
          isEmpty: isNullOrEmpty,
          child: child);
    }
    final hasBorder = backgroundStyle.borderWidth != null;

    if (hasBorder) {
      return InputDecorator(
          decoration: effectiveDecoration.copyWith(
              prefixIcon: prefixIcon,
              errorText: field.errorText,
              isDense: isDense,
              contentPadding: contentPadding,
              hintText: hintText,
              border: OutlineInputBorder(
                  borderSide: BorderSide(
                      color: backgroundStyle.borderColor ?? Color(0xFFEEEEEE),
                      width: backgroundStyle.borderWidth ?? 1))),
          isEmpty: isNullOrEmpty,
          child: child);
    } else {
      return InputDecorator(
          decoration: effectiveDecoration.copyWith(
              hintText: hintText, errorText: field.errorText, isDense: isDense),
          isEmpty: isNullOrEmpty,
          child: Container(
              constraints: BoxConstraints(minHeight: 48),
              color: backgroundStyle.backgroudColor,
              padding: contentPadding,
              child: child));
    }
  }

  @override
  FormInputStateBase<T> createState() => FormInputStateBase<T>();
}

class FormInputStateBase<T> extends FormFieldState<T> {
  ValueNotifier<T?>? _valueController;

  ValueNotifier<T?>? get effectiveValueController =>
      widget.valueController ?? _valueController;

  @override
  FormInputBase<T> get widget => super.widget as FormInputBase<T>;

  @override
  void initState() {
    super.initState();
    if (widget.valueController == null) {
      _valueController = ValueNotifier<T?>(widget.initialValue);
    } else {
      widget.valueController!.addListener(_handleControllerChanged);
    }
  }

  @override
  void didUpdateWidget(FormInputBase<T> oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.valueController != oldWidget.valueController) {
      oldWidget.valueController?.removeListener(_handleControllerChanged);
      widget.valueController?.addListener(_handleControllerChanged);

      if (oldWidget.valueController != null && widget.valueController == null) {
        _valueController = ValueNotifier<T>(oldWidget.valueController!.value);
      }
      if (widget.valueController != null) {
        setValue(widget.valueController!.value);
        if (oldWidget.valueController == null) _valueController = null;
      }
    }
  }

  @override
  void dispose() {
    widget.valueController?.removeListener(_handleControllerChanged);
    super.dispose();
  }

  @override
  void reset() {
    super.reset();
    setState(() {
      effectiveValueController!.value = widget.initialValue;
    });
  }

  void _handleControllerChanged() {
    changeValue(effectiveValueController!.value);
  }

  void changeValue(T? v) {
    if (effectiveValueController!.value != v) {
      effectiveValueController!.value = v;
    }
    if (v != value) {
      didChange(v);
    }
  }
}

class ValueController<T> extends ValueNotifier<T> {
  ValueController(T value) : super(value);

  ValueController.fromValue(T value) : super(value);
}
